@@ -1,5 +1,5 @@
 #!/usr/bin/python2.5
-import sys, os, gtk, gobject, hildon, dbus, signal, alarm, time
+import sys, os, fcntl, gtk, gobject, hildon, dbus, signal, alarm, time
 import pyconic; conic=pyconic # ideally, import conic, but that's broken
 import syncevolution
 
@@ -82,17 +82,7 @@ class SyncRunner(object):
     url = config.get("syncURL")
     self.method = get_method(url)
     # grab exclusive lock
-    while True:
-      try:
-        fd = os.open(self.dir + "/.sync_lock", os.O_WRONLY|os.O_CREAT|os.O_EXCL)
-        break
-      except:
-        # FIXME: make sure the old process isn't dead before returning
-        # (a bit of work to do without race conditions...)
-        return
-    pid = os.getpid()
-    os.write(fd, "%d" % pid)
-    os.close(fd)
+    self.acquire_lock()
     # connect to network
     self.connection = conic.Connection()
     flags = conic.CONNECT_FLAG_NONE
@@ -100,6 +90,46 @@ class SyncRunner(object):
       flags |= conic.CONNECT_FLAG_AUTOMATICALLY_TRIGGERED
     self.connection.connect("connection-event", self.connected)
     self.connection.request_connection(flags)
+  def acquire_lock(self):
+    lock_name = self.dir + "/.sync_lock"
+    while True:
+      try:
+        fd = os.open(lock_name, os.O_RDWR|os.O_CREAT|os.O_EXCL)
+        if fcntl.flock(fd, fcntl.LOCK_EX) < 0:
+          # since we're blocking, this shouldn't happen, in theory
+          os.close(fd)
+          continue
+        break
+      except:
+        # check if this might be a stale lock
+        try:
+          fd = os.open(lock_name, os.O_RDWR)
+        except:
+          # file vanished, try to lock again
+          continue
+        if fcntl.flock(fd, fcntl.LOCK_EX) < 0:
+          # since we're blocking, this shouldn't happen, in theory
+          os.close(fd)
+          continue
+        buf = os.read(fd)
+        if buf == "":
+          # another process might be in the process of locking
+          # (opened the file but didn't lock it yet)
+          os.close(fd)
+          continue
+        pid = int(buf)
+        # since the file isn't empty, and we have the lock,
+        # the other process must have terminated prematurely.
+        # Take over the lock.
+        os.lseek(fd, 0, 0)
+        os.ftruncate(fd, 0)
+        break
+    self.lock_fd = fd
+    pid = os.getpid()
+    os.write(fd, "%d" % pid)
+  def release_lock(self):
+    os.unlink(self.dir + "/.sync_lock")
+    os.close(self.lock_fd)
   def connected(self, connection, event):
     self.conn_status = event.get_status()
     if self.conn_status == conic.STATUS_CONNECTED:
@@ -177,13 +207,13 @@ class SyncRunner(object):
         status["result"] = RESULT_LOCKED
       else:
         status["result"] = RESULT_NO_CONNECTION
-        os.unlink(self.dir + "/.sync_lock")
+        self.release_lock()
       return status
     self.deactivate_watches()
     status = self.sync.synchronize_status(self.state)
     self.connection.disconnect()
     self.state = None
-    os.unlink(self.dir + "/.sync_lock")
+    self.release_lock()
     return status
 
 class SyncGUI(object):
@@ -312,7 +342,7 @@ class SyncGUI(object):
     window.destroy()
   def do_delete_server(self, server):
     server_dir = self.sync.get_server_dir(server)
-    server_cookie = load_server_cookie(server_dir)
+    server_cookie = self.load_server_cookie(server_dir)
     if not server_cookie is None:
       os.unlink(server_dir + "/.alarmcookie")
       alarm.delete_event(server_cookie)
